"""
@author: Thilogen
"""

import numpy as np
import matplotlib.pyplot as plt

#-------------------------------------------------------------------

# Define some functions needed to calculate the necessary equations

def skewm(r):
    return np.array([[0,-r[2],r[1]], [r[2],0,-r[0]], [-r[1],r[0],0]])
def vex(u):
    return np.array([u[2,1], u[0,2], u[1,0]])
def expso3(u):
    S = skewm(u); un = np.linalg.norm(u)
    return np.identity(3) + np.sinc(un/np.pi)*S + 0.5*(np.sinc(un/(2*np.pi)))**2 * S@S

def expse2_3(u):
    un = np.linalg.norm(u[0:3]) 
    if un > 0.000001:
        a = ((1-np.cos(un)) / un**2)
        b = (un - np.sin(un))/(un**3)
    else:
        a = 1/2 + (un**2)/24
        b = 1/6 + (un**2)/120       
    S = np.block([[skewm(u[0:3]), u[3:6].reshape(3,1), u[6:9].reshape(3,1)],
                  [np.zeros((2,5))]])
    return np.identity(5) + S + a*S@S + b*S@S@S

def integrate_se3(h, X, om, u):
    R = X[0:3,0:3]; v = X[0:3,3]; x = X[0:3,4]
    R_int = R @ expso3(h*om)
    v = np.array([1.0, 0.0, 0.0])
    x_int = x + h * R @ JLso3(h*om) @ v  
    return np.hstack((R_int, v.reshape(3,1), x_int.reshape(3,1)))

def JLso3(u):
    theta = np.linalg.norm(u); S = skewm(u)
    a = 0.5 * (np.sinc(theta/(2*np.pi)))**2
    if theta > 0.000001:
        b = (theta - np.sin(theta))/(theta**3)
    else:
        b = 1/6 + theta**2/120
    return np.eye(3) + a*S + b*S@S

def logSO3(R):
    # The vector form of the logarithm in SO(3) (Robotic style)
    theta = np.arccos(0.5 * (np.trace(R) - 1))
    ct = np.cos(theta); vt = 1-ct
    if theta < 0.000001:
        f = 1 + (theta**2)/6
        u = vex(f * 0.5 * (R-R.T))
    if np.pi - theta < 0.00001:
        ct = np.cos(theta); vt = 1-ct
        if R[0,0] - ct > 0.5:
            kx = np.sqrt((R[0,0] - ct)/vt)
            ky = (R[0,1] + R[1,0]) / (2*kx*vt)
            kz = (R[2,0] + R[0,2]) / (2*kx*vt)
        elif R[1,1] - ct > 0.5:
            ky = np.sqrt((R[1,1] - ct)/vt)
            kz = (R[1,2] + R[2,1]) / (2*ky*vt)
            kx = (R[0,1] + R[1,0]) / (2*ky*vt)
        else:
            kz = np.sqrt((R[2,2] - ct)/vt)
            kx = (R[2,0] + R[0,2]) / (2*kz*vt)
            ky = (R[1,2] + R[2,1]) / (2*kz*vt)
            u = theta * np.array([kx, ky, kz])
    else:
        f = theta / np.sin(theta)
        u = vex(f * 0.5 * (R-R.T))
    return u

#-------------------------------------------------------------------

# Define the function that creates the real trajectory

def create_dynamics(n_time, h, om, v, u):
    X = np.zeros((3,5,n_time))
    X[:,0:3, 0] = np.eye(3)
    X[:,3, 0] = v
    for i in range(1,n_time):
        X[:,:,i] = integrate_se3(h, X[:,:,i-1].copy(), om, u)     
    return X

def generate_map(v, omega, delta_r, n_L):
    p_L = np.zeros((3,n_L)); r = v/omega
    for i in range(n_L):
        rho = r - 2*delta_r
        theta = 2*np.pi*i/n_L
        p_L[:,i] = [rho*np.cos(theta), rho*np.sin(theta)+5.0, 0.8*((-1)**i)] 
    return p_L

#-------------------------------------------------------------------

#Define the propagate function for both filter 

def propagate(Xh, P, h, om, u, g, Q, config):
    Rh = Xh[:, 0:3]; vh = Xh[:,3]; xh = Xh[:,4]
    
    zero_3 = np.zeros((3,3))
    A = np.zeros((9,9))
    if config == "mekf":
        A[3:6,0:3] = -skewm(Rh @ u); A[6:9,3:6] = np.eye(3)
    elif config == "riekf":
        A[3:6,0:3] = skewm(g); A[6:9,3:6] = np.eye(3)   
        
        
    G = np.block([[Rh, zero_3, zero_3],
          [skewm(vh), Rh, zero_3],
          [skewm(xh), zero_3, Rh]])
    
    Q_hat = G @ Q @ G.T
     
    P_p = P + h * (A @ P + P @ A.T + Q_hat)
    
    
    if config == "mekf":
        Xh_p = integrate_se3(h, Xh, om, u)
    elif config == "riekf":
        Xh_p = np.block([[integrate_se3(h, Xh, om, u)],
                         [np.zeros((2,3)), np.eye(2)]])
        

    return Xh_p, P_p

#Define the update function for both filter 

def update(Xh_p, P_p, Y, N, config):
    Rh_p = Xh_p[0:3,0:3]; vh_p = Xh_p[0:3,3]; xh_p = Xh_p[0:3,4]
    R = Y[0:3,0:3]; x = Y[0:3,4]
    
    yn = np.zeros(3 * n_L) # 1x9 - vec
    yn_h = np.zeros(3 * n_L) # 1x9 - vec
    
    for i in range(n_L):
        yn[3*i:3*i+3] = R.T @ (p_L[:,i] - x)
        yn_h[3*i:3*i+3] = Rh_p.T @ (p_L[:,i] - xh_p)

    N_hat = np.zeros((3*n_L,3*n_L))
    H = np.zeros((3*n_L,3*n_L))
    
    for i in range(n_L):
        N_hat[3*i:3*i+3, 3*i:3*i+3] = Rh_p @ N[3*i:3*i+3, 3*i:3*i+3] @ Rh_p.T
        
 
    if config == "mekf":
        for i in range(n_L):
            H[3*i:3*i+3, 0:3] = Rh_p.T @ skewm(p_L[:,i] - xh_p)
            H[3*i:3*i+3, 6:9] = -Rh_p.T

    elif config == "riekf":
        for i in range(n_L):
            H[3*i:3*i+3, 0:3] = skewm(p_L[:,i])
            H[3*i:3*i+3, 6:9] = -np.eye(3)

    S = H @ P_p @ H.T + N_hat # 9x9 - mat
    L_n = P_p @ H.T @ np.linalg.inv(S) # 9x9 - mat
    
    if config == "mekf":
        inno =  -L_n @ (yn_h - yn) # 9x1 vec
    
        Xh_u = np.zeros((3,5))
        Xh_u[:,0:3] = Xh_p[:,0:3] @ expso3(inno[0:3])
        Xh_u[:,3] = Xh_p[:,3] + inno[3:6] 
        Xh_u[:,4] = Xh_p[:,4] + inno[6:9]

    elif config == "riekf":
        yn_tilde = np.zeros(3 * n_L) # 1x9 - vec
        for i in range(n_L):
            yn_tilde[3*i:3*i+3] = Rh_p @ (yn_h[3*i:3*i+3] - yn[3*i:3*i+3]) # 1x3 - vec
                   
        delta = -L_n @ yn_tilde # 9x1 - mat
            
        Xhi_u = expse2_3(delta) @ Xh_p
        Xh_u = Xhi_u[0:3,:]

            

    P_u = (np.eye(9) - L_n @ H) @ P_p


    return Xh_u, P_u

#----------------------------------------------*---------------------

# Define the system initialization parameters 

radius = 5.0; t_circ = 30.0; omega = 2*np.pi/t_circ; v_t = omega * radius
h = 0.1
n_time = 1*int(t_circ/h)

n_L = 3

ex = np.array([1,0,0]); ey = np.array([0,1,0]); ez = np.array([0,0,1])
g =  -9.81 * ez
v_3d = v_t * ex; om_3d = omega * ez; u_3d = omega**2 * radius * ey + g 

#-------------------------------------------------------------------

init_heading_est_error =  -15 * np.pi/180
theta_t_i = init_heading_est_error

# Define the noise intialization parameters

I_3 = np.eye(3)
zero_3 = np.zeros((3,3))

N = (1/h) * 1.0e-2 * np.block([[I_3, zero_3, zero_3],
                               [zero_3, I_3, zero_3],
                               [zero_3, zero_3, I_3]])

Q1 = 1.0e-8 * np.block([[I_3, zero_3, zero_3],
                     [zero_3, I_3, zero_3],
                     [zero_3, zero_3, zero_3]])


Q2 = 1.0e-4 * np.block([[I_3, zero_3, zero_3],
                     [zero_3, I_3, zero_3],
                     [zero_3, zero_3, zero_3]])


Q10 =  np.block([[((5*np.pi/180)**2)*I_3, zero_3, zero_3],
                 [zero_3, 1e-2*I_3, zero_3],
                 [zero_3, zero_3, I_3]])


Q20 =  np.block([[((15*np.pi/180)**2)*I_3, zero_3, zero_3],
                 [zero_3, 1e-2*I_3, zero_3],
                 [zero_3, zero_3, I_3]])



#-------------------------------------------------------------------

# Create the true trajectory and true landmark positions

X = create_dynamics(n_time, h, om_3d, v_3d, u_3d)
p_L = generate_map(v_t, omega, 1, n_L)

filters = ["mekf","riekf"]

cov_om = Q2

Xu_mekf = np.zeros((3,5, n_time))
Xu_riekf = np.zeros((3,5, n_time))

Rt_init = expso3(np.array([0,0, init_heading_est_error]))

#Attitude
Xu_mekf[:,0:3,0] = Rt_init 
Xu_riekf[:,0:3,0] = Rt_init

#Velocity
Xu_mekf[:,3,0] = Rt_init @ v_3d
Xu_riekf[:,3,0] = Rt_init @ v_3d

#Position
Xu_mekf[:,4,0] = np.array([1.0, 0.0, 1.0])
Xu_riekf[:,4,0] = np.array([1.0, 0.0, 1.0])

Pu_riekf = np.zeros((9, 9, n_time))
Pu_mekf = np.zeros((9, 9, n_time))

if np.array_equal(cov_om, Q1):
    Pu_riekf[:,:,0] = Q10
    Pu_mekf[:,:,0] = Q10

elif np.array_equal(cov_om, Q2):
    Pu_riekf[:,:,0] = Q20
    Pu_mekf[:,:,0] = Q20

#-------------------------------------------------------------------

plt.figure(1)
plt.figure(1).clear()
ax = plt.axes(projection='3d')

ax.plot3D(p_L[0,:], p_L[1,:], p_L[2,:],"ko", label = "Landmarks")
ax.plot3D(p_L[0,:], p_L[1,:], np.zeros(3),"kx")
for i in range(n_L):
    px = np.hstack((p_L[0,i], p_L[0,i]))
    py = np.hstack((p_L[1,i], p_L[1,i]))
    pz = np.hstack((p_L[2,i], 0))
    ax.plot3D(px, py, pz, "k--")
    
ax.plot3D(X[0,4,:], X[1,4,:], X[2,4,:],"g", label = "True trajectory")

#Iterate through the filter list and calculate and plot trajectory for each filter

for config in filters:
    for k in range(1, n_time):
        Y = X[:,:,k]
        if config == "mekf":
            Xp, Pp = propagate(Xu_mekf[:,:,k-1].copy(), Pu_mekf[:,:,k-1].copy(), 
                                                            h, om_3d, u_3d, g, cov_om, config)
            Xu_mekf[:,:,k], Pu_mekf[:,:,k] = update(Xp, Pp, Y, N, config)
        elif config == "riekf":
            Xp, Pp = propagate(Xu_riekf[:,:,k-1].copy(), Pu_riekf[:,:,k-1].copy(), 
                                                            h, om_3d, u_3d, g, cov_om, config)
            Xu_riekf[:,:,k], Pu_riekf[:,:,k] = update(Xp, Pp, Y, N, config)
            
    if config == "mekf":
        ax.plot3D(Xu_mekf[0,4,:], Xu_mekf[1,4,:], Xu_mekf[2,4,:], "b--", label = "MEKF - Estimated trajectory")
 
    
    elif config == "riekf":
        ax.plot3D(Xu_riekf[0,4,:], Xu_riekf[1,4,:], Xu_riekf[2,4,:],"r--", label = "RIEKF - Estimated trajectory")
    

ax.set_zlim(-1,5)
plt.legend()


if np.array_equal(cov_om, Q1):
    plt.title("Pose estimation, Q = Q1")
    plt.savefig("navigation_pose_estimation_Q1.png")
elif np.array_equal(cov_om, Q2):
    plt.title("Pose estimation, Q = Q2")
    plt.savefig("navigation_pose_estimation_Q2.png")


plt.show()

#-------------------------------------------------------------------

#Plot the attitude error

plt.figure(2)
plt.figure(2).clear()

t = h*np.arange(n_time)


riekf_attitude_error = np.zeros(n_time)
mekf_attitude_error = np.zeros(n_time)

for i in range(n_time):
    riekf_attitude_error[i] = abs(logSO3(X[0:3,0:3,i] @ Xu_riekf[0:3,0:3,i].T).sum())
    mekf_attitude_error[i] = abs(logSO3(X[0:3,0:3,i] @ Xu_mekf[0:3,0:3,i].T).sum())
    
    
    
plt.plot(t, riekf_attitude_error / (np.pi/180), "r--", label = "RIEKF attitude error")
plt.plot(t, mekf_attitude_error / (np.pi/180), "b--", label = "MEKF attitude error")



plt.xlim([0, n_time*h])
plt.ylim([0, 20])
plt.legend()
plt.xlabel("time [s]")
plt.ylabel("Attitude error [degrees]")

if np.array_equal(cov_om, Q1):
    plt.title("Attitude error, Q = Q1")
    plt.savefig("navigation_attitude_error_Q1.png")
elif np.array_equal(cov_om, Q2):
    plt.title("Attitude error, Q = Q2")
    plt.savefig("navigation_attitude_error_Q2.png")
    
plt.show()

#-------------------------------------------------------------------

#Plot the position error

plt.figure(3)
plt.figure(3).clear()

riekf_position_error = np.zeros(n_time)
mekf_position_error = np.zeros(n_time)

for i in range(n_time):
    riekf_position_error[i] = np.linalg.norm(X[0:3,4,i] - Xu_riekf[0:3,4,i])
    mekf_position_error[i] = np.linalg.norm(X[0:3,4,i] - Xu_mekf[0:3,4,i])
    

plt.plot(t, riekf_position_error, "r--", label = "RIEKF position error")
plt.plot(t, mekf_position_error, "b--", label = "MEKF position error")
plt.xlim([0, n_time*h])
plt.ylim([0, 2])
plt.xlabel("time [s]")
plt.ylabel("Position error [m]")
plt.legend()
if np.array_equal(cov_om, Q1):
    plt.title("Position error, Q = Q1")
    plt.savefig("navigation_position_error_Q1.png")
elif np.array_equal(cov_om, Q2):
    plt.title("Position error, Q = Q2")
    plt.savefig("navigation_position_error_Q2.png")
plt.show()




